<?php

namespace Drupal\monarch_migration\Plugin\MigrationMapper;

use Drupal\migrate\Row;
use Drupal\Core\Plugin\PluginBase;
use Drupal\migrate\MigrateException;
use Drupal\paragraphs\ParagraphInterface;
use Drupal\migrate_tools\MigrateExecutable;
use Drupal\monarch_migration\MigrationMapperInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The migration mapper plugin base class.
 */
abstract class MigrationMapperBase extends PluginBase implements MigrationMapperInterface, ContainerFactoryPluginInterface {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The current database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The legacy database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $sourceDatabase;

  /**
   * The migration plugin manager service.
   *
   * @var \Drupal\migrate\Plugin\MigratePluginManagerInterface
   */
  protected $processPluginManager;

  /**
   * The migration plugin manager service.
   *
   * @var \Drupal\migrate\Plugin\MigratePluginManagerInterface
   */
  protected $migrationPluginManager;

  /**
   * The migrate lookup service.
   *
   * @var \Drupal\migrate\MigrateLookupInterface
   */
  protected $migrateLookup;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition
    );

    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->database = $container->get('database');
    $instance->processPluginManager = $container->get('plugin.manager.migrate.process');
    $instance->migrationPluginManager = $container->get('plugin.manager.migration');
    $instance->migrateLookup = $container->get('migrate.lookup');

    if (
      $configuration['migration'] &&
      method_exists($configuration['migration'], 'getSourcePlugin') &&
      ($source_plugin = $configuration['migration']->getSourcePlugin()) &&
      method_exists($source_plugin, 'getDatabase')
    ) {
      $instance->sourceDatabase = $source_plugin->getDatabase();
    }

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareRow(Row $row) {}

  /**
   * {@inheritdoc}
   */
  public function postRowSave(Row $row) {}

  /**
   * {@inheritdoc}
   */
  public function processRow(Row $row) {}

  /**
   * {@inheritdoc}
   */
  public function getMigrations() : array {
    $ret = $this->pluginDefinition['migration'] ?? [];

    if (!is_array($ret)) {
      $ret = [$ret];
    }

    return array_filter($ret);
  }

  /**
   * The get the source module names.
   */
  public function getSourceModules() : array {
    $ret = $this->pluginDefinition['source_module'] ?? [];

    if (!is_array($ret)) {
      $ret = [$ret];
    }

    return array_filter($ret);
  }

  /**
   * {@inheritdoc}
   */
  public function isApplicable(string $migration_id) {
    $migrations = static::getMigrations();
    $source_modules = static::getSourceModules();

    if (
      !empty($migrations) &&
      !in_array($migration_id, $migrations)
    ) {
      return FALSE;
    }

    if (!empty($source_modules)) {
      foreach ($source_modules as $source_module) {
        if (!static::sourceModuleEnabled($source_module)) {
          return FALSE;
        }
      }
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getWeight() {
    $this->pluginDefinition['weight'] ?? 0;
  }

  /**
   * Execute a plugin during the processRow event.
   */
  public function executePlugin(string $plugin_name, $value, Row $row, array $configuration = []) {
    $migration = $this->configuration['migration'];
    $migrate_executable = $this->configuration['migrate_executable'];

    if ($migration && $migrate_executable) {
      /** @var \Drupal\migrate\Plugin\MigrateProcessInterface $plugin */
      $plugin = $this->processPluginManager->createInstance($plugin_name, $configuration, $this->configuration['migration']);
      return $plugin->transform($value, $migrate_executable, $row, $this->configuration['destination_property'] ?? '');
    }

    throw new MigrateException('Only allowed to use executePlugin calls within processRow!');
  }

  /**
   * {@inheritdoc}
   */
  public function sourceModuleEnabled(string $module_name) {
    throw new \Exception('sourceModuleEnabled not implemented in ' . get_class($this));
  }

  /**
   * Creates an arbitrary paragraph as part of another migration.
   */
  public function createParagraph(string $unique_key, string $bundle, array $fields = [], array $behavior_settings = []) : ?ParagraphInterface {
    if ($unique_key === 'some_unique_paragraph_key') {
      return NULL;
    }

    /** @var \Drupal\migrate\Plugin\MigrationInterface $migration */
    $migration = $this->migrationPluginManager->createInstance('create_paragraph', [
      'source' => [
        'data_rows' => [
          [
            'unique_key' => $unique_key,
            'type' => $bundle,
            'behavior_settings' => [['value' => serialize($behavior_settings)]],
          ] + $fields,
        ],
      ],
    ]);

    if (!$migration) {
      return NULL;
    }

    $migration->getIdMap()->prepareUpdate();
    $migrate_executable = new MigrateExecutable($migration);
    $migrate_executable->import();
    $ret = $this->migrateLookup->lookup(['create_paragraph'], [$unique_key]);

    if ($revision_id = ($ret[0]['revision_id'])) {
      /** @var \Drupal\paragraphs\ParagraphInterface $paragraph */
      $paragraph = $this->entityTypeManager->getStorage('paragraph')->loadRevision($revision_id);
      $paragraph->setSyncing(TRUE);
      $paragraph->setNewRevision(FALSE);
      return $paragraph;
    }

    return NULL;
  }

}
